@@ -6,13 +6,18 @@ from quickstats.utils.common_utils import get_class_that_defined_method
 import ROOT
 
 class AnalysisObject(AbstractObject):
+    
+    kInitialSnapshotName = "initialSnapshot"
+    kCurrentSnapshotName = "currentSnapshot"
+    kTempSnapshotName    = "tmpSnapshot"
+    
     def __init__(self, filename:Optional[str]=None, poi_name:Optional[Union[str, List[str]]]=None,
                  data_name:str='combData', binned_likelihood:bool=True,
                  fix_param:str='', profile_param:str='', ws_name:Optional[str]=None, 
                  mc_name:Optional[str]=None, snapshot_name:Optional[Union[List[str], str]]=None,
                  minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', precision:float=0.001, 
                  eps:float=1.0, retry:int=1, strategy:int=1, print_level:int=-1, timer:bool=False,
-                 num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False,
+                 num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False, hesse:bool=False,
                  fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                  max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
                  int_bin_precision:float=-1., preset_param:bool=False,
@@ -25,13 +30,13 @@ class AnalysisObject(AbstractObject):
         if filename is not None:
             self.setup_model(filename, data_name, binned_likelihood, ws_name, mc_name, 
                              snapshot_name, fix_cache, fix_multi, verbosity=verbosity)
-            self.save_snapshot("initialSnapshot")
+            self.save_snapshot(self.kInitialSnapshotName)
             self.set_poi(poi_name)
             if preset_param:
                 self.preset_parameters()
             self.setup_parameters(fix_param, profile_param, update_snapshot=True)
             self.setup_minimizer(minimizer_type, minimizer_algo, precision, eps, retry, strategy,
-                                 num_cpu, offset, optimize, eigen, max_calls, max_iters, print_level,
+                                 num_cpu, offset, optimize, eigen, hesse, max_calls, max_iters, print_level,
                                  timer, constrain_nuis, batch_mode, int_bin_precision,
                                  verbosity=verbosity)
             self.sanity_check()
@@ -110,19 +115,45 @@ class AnalysisObject(AbstractObject):
     def setup_parameters(self, fix_param:str='', profile_param:str='', update_snapshot:bool=True):
         if not self.model:
             raise RuntimeError('uninitialized analysis object')
+        fixed_parameters = []
+        profiled_parameters = []
         if fix_param:
-            self.model.fix_parameters(fix_param)
+            fixed_parameters = self.model.fix_parameters(fix_param)
         if profile_param:
-            self.model.profile_parameters(profile_param)
+            profiled_parameters = self.model.profile_parameters(profile_param)
         if update_snapshot:
-            self.save_snapshot("currentSnapshot")
+            self.save_snapshot(self.kCurrentSnapshotName)
+        self._check_poi_setup(fixed_parameters, profiled_parameters)
+    
+    def _check_poi_setup(self, fixed_parameters:List, profiled_parameters:List):
+        # pois defined in workspace
+        ws_pois  = self.model.pois
+        # pois defined in this analysis
+        ana_pois = ROOT.RooArgSet(self.poi)
+        if ana_pois.size() == 0:
+            return None
+        aux_pois = ROOT.RooArgSet(ws_pois)
+        aux_pois.remove(ana_pois)
+        for profiled_param in profiled_parameters:
+            if aux_pois.contains(profiled_param):
+                param_name = profiled_param.GetName()
+                self.stdout.warning(f"WARNING: Attemping to profile the parameter \"{param_name}\" which is a "
+                                    f"POI defined in the workspace but not a POI used in this study. This "
+                                    f"parameter will still be fixed during likelihood fit / limit setting. "
+                                    f"Please designate the parameter as a POI in this study if intended.", "red")
+        for fixed_param in fixed_parameters:
+            if ana_pois.contains(fixed_param):
+                param_name = fixed_param.GetName()
+                self.stdout.warning(f"WARNING: Attemping to fix the parameter \"{param_name}\" which is a "
+                                    f"POI used in this study. This parameter will still be floated during "
+                                    f"unconditional likelihood fit / limit setting.", "red")
             
     def setup_minimizer(self, minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                         precision:float=0.001, eps:float=1.0, retry:int=1, strategy:int=0, 
                         num_cpu:int=1, offset:bool=True, optimize:bool=True, eigen:bool=False,
-                        max_calls:int=-1, max_iters:int=-1, print_level:int=-1, timer:bool=False,
-                        constrain_nuis:bool=True, batch_mode:bool=False, int_bin_precision:float=-1.,
-                        verbosity:Optional[Union[int, str]]="INFO"):
+                        hesse:bool=False, max_calls:int=-1, max_iters:int=-1, print_level:int=-1,
+                        timer:bool=False, constrain_nuis:bool=True, batch_mode:bool=False,
+                        int_bin_precision:float=-1., verbosity:Optional[Union[int, str]]="INFO"):
         
         minimizer = ExtendedMinimizer("Minimizer", self.model.pdf, self.model.data, verbosity=verbosity)
         
@@ -149,6 +180,7 @@ class AnalysisObject(AbstractObject):
             'eps'              : eps,
             'retry'            : retry,
             'eigen'            : eigen,
+            'hesse'            : hesse,
             'max_calls'        : max_calls,
             'max_iters'        : max_iters,
             'print_level'      : print_level,
