@@ -20,13 +20,16 @@ from quickstats.components import AbstractObject
 from quickstats.maths.numerics import is_integer, pretty_value
 from quickstats.utils.root_utils import load_macro
 from quickstats.utils.common_utils import str_list_filter
-from quickstats.utils.roofit_utils import get_variable_attributes, variable_collection_to_dataframe
-
+from quickstats.utils.roofit_utils import (get_variable_attributes, variable_collection_to_dataframe,
+                                           get_relevant_components)
 from quickstats.interface.root import TH1, RooDataSet, RooArgSet, RooMsgService
+from quickstats.interface.root import roofit_extension as rf_ext
 from quickstats.components.basics import WSVariables, SetValueMode
 
 class ExtendedModel(AbstractObject):
     
+    _DEFAULT_CONSTR_CLS_ = ["RooGaussian", "RooLognormal", "RooGamma", "RooPoisson", "RooBifurGauss"]
+    
     _DEFAULT_NAMES_ = {
         'conditional_globs': 'conditionalGlobs_{mu}',
         'conditional_nuis': 'conditionalNuis_{mu}',
@@ -61,6 +64,7 @@ class ExtendedModel(AbstractObject):
         self.fix_multi = fix_multi
         self.interpolation_code = interpolation_code
         
+        quickstats.load_corelib()
         if load_extension:
             self.load_extension()
         
@@ -119,10 +123,7 @@ class ExtendedModel(AbstractObject):
     
     @semistaticmethod
     def load_extension(self):
-        if (quickstats.root_version >= (6, 26, 0)):
-            extensions = ['RooTwoSidedCBShape']
-        else:
-            extensions = ['RooTwoSidedCBShape', 'FlexibleInterpVarMkII']
+        extensions = quickstats.get_workspace_extensions()
         for extension in extensions:
             result = load_macro(extension)
             if (result is not None) and hasattr(ROOT, extension):
@@ -191,42 +192,43 @@ class ExtendedModel(AbstractObject):
     def initialize(self):
         if isinstance(self.fname, str):
             if not os.path.exists(self.fname):
-                raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
-            self.stdout.info('INFO: Opening file "{}"'.format(self.fname))
+                raise FileNotFoundError(f'workspace file "{self.fname}" does not exist')
+            self.stdout.info(f'INFO: Opening file "{self.fname}"')
             file = ROOT.TFile(self.fname) 
             if (not file):
-                raise RuntimeError("Something went wrong while loading the root file: {}".format(self.fname))
+                raise RuntimeError(f"Something went wrong while loading the root file: {self.fname}")
             # load workspace
             if self.ws_name is None:
                 ws_names = [i.GetName() for i in file.GetListOfKeys() if i.GetClassName() == 'RooWorkspace']
                 if not ws_names:
-                    raise RuntimeError("No workspaces found in the root file: {}".format(self.fname))
+                    raise RuntimeError(f"No workspaces found in the root file: {self.fname}")
                 if len(ws_names) > 1:
-                    self.stdout.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
-                          " are \"{}\". Will choose the first one by default".format(self.fname, ','.join(ws_names)))
+                    self.stdout.warning("WARNING: Found multiple workspace instances from the root file: "
+                                        f"{self.fname}. Available workspaces are \"{','.join(ws_names)}\". "
+                                        f"Will choose the first one by default")
                 self.ws_name = ws_names[0]
             ws = file.Get(self.ws_name)
         elif isinstance(self.fname, ROOT.RooWorkspace):
             file = None
             ws = self.fname
         if not ws:
-            raise RuntimeError('failed to load workspace "{}"'.format(self.ws_name))
+            raise RuntimeError(f'failed to load workspace "{self.ws_name}"')
         self.ws_name = ws.GetName()
-        self.stdout.info('INFO: Loaded workspace "{}"'.format(self.ws_name))
+        self.stdout.info(f'INFO: Loaded workspace "{self.ws_name}"')
         # load model config
         if self.mc_name is None:
             mc_names = [i.GetName() for i in ws.allGenericObjects() if 'ModelConfig' in i.ClassName()]
             if not mc_names:
-                raise RuntimeError("no ModelConfig object found in the workspace: {}".format(ws_name))
+                raise RuntimeError(f"no ModelConfig object found in the workspace: {ws_name}")
             if len(mc_names) > 1:
-                self.stdout.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
-                      "Available ModelConfigs are \"{}\". "
-                      "Will choose the first one by default".format(ws_name, ','.join(mc_names)))
+                self.stdout.warning(f"WARNING: Found multiple ModelConfig instances from the workspace: {ws_name}. "
+                                    f"Available ModelConfigs are \"{','.join(mc_names)}\". "
+                                    "Will choose the first one by default")
             self.mc_name = mc_names[0]
         model_config = ws.obj(self.mc_name)
         if not model_config:
-            raise RuntimeError('failed to load model config "{}"'.format(self.mc_name))
-        self.stdout.info('INFO: Loaded model config "{}"'.format(self.mc_name))
+            raise RuntimeError(f'failed to load model config "{self.mc_name}"')
+        self.stdout.info(f'INFO: Loaded model config "{self.mc_name}"')
             
         # modify interpolation code
         if self.interpolation_code != -1:
@@ -249,13 +251,13 @@ class ExtendedModel(AbstractObject):
         pdf = model_config.GetPdf()
         if not pdf:
             raise RuntimeError('Failed to load pdf')
-        self.stdout.info('INFO: Loaded model pdf "{}" from model config'.format(pdf.GetName()))
+        self.stdout.info(f'INFO: Loaded model pdf "{pdf.GetName()}" from model config')
              
         # load dataset
         if self.data_name is None:
             data_names = [i.GetName() for i in ws.allData()]
             if not data_names:
-                raise RuntimeError("no datasets found in the workspace: {}".format(ws_name))
+                raise RuntimeError(f"no datasets found in the workspace: {ws_name}")
             self.data_name = data_names[0]
         data = ws.data(self.data_name)
         if not data:
@@ -376,6 +378,8 @@ class ExtendedModel(AbstractObject):
                 v_min, v_max = value[0], value[1]
             else:
                 raise ValueError('invalid expression for profiling parameter: {}'.format(value))
+            if new_value is None:
+                new_value = old_value
             # set range
             if (v_min is not None) and (v_max is not None):
                 if (new_value < v_min) or (new_value > v_max):
@@ -472,145 +476,26 @@ class ExtendedModel(AbstractObject):
         pseudo_globs = pdf.generateSimGlobal(globs, 1)
         pdf_vars = pdf.getVariables()
         pdf_vars.assignValueOnly(pseudo_globs.get(0))
-    
-    @staticmethod
-    def find_unique_prod_components(root_pdf, components, recursion_count=0):
-        if (recursion_count > 50):
-            raise RuntimeError('find_unique_prod_components detected infinite loop')
-        pdf_list = root_pdf.pdfList()
-        if pdf_list.getSize() == 1:
-            components.add(pdf_list)
-            #print('ProdPdf {} is fundamental'.format(pdf_list.at(0).GetName()))
-        else:
-            for pdf in pdf_list:
-                if pdf.ClassName() != 'RooProdPdf':
-                    #print('Pdf {} is no RooProdPdf. Adding it.')
-                    components.add(pdf)
-                    continue
-                find_unique_prod_components(pdf, components, recursion_count+1)
-                
-    @staticmethod 
-    def _unfold_constraints(source:ROOT.RooArgSet, obs:ROOT.RooArgSet, nuis:ROOT.RooArgSet, 
-                           result:ROOT.RooArgSet=None, recursion:int=0, threshold:int=50):
-        if recursion > threshold:
-            raise RuntimeError("failed to find unfold constraints: recusion limit exceeded")
-        if result is None:
-            result = ROOT.RooArgSet()
-        for pdf in source:
-            class_name = pdf.ClassName()
-            if class_name not in ["RooGaussian", "RooLognormal", "RooGamma", "RooPoisson", "RooBifurGauss"]:
-                constraint_set = pdf.getAllConstraints(obs.Clone(), nuis.Clone(), ROOT.kFALSE)
-                ExtendedModel._unfold_constraints(constraint_set, obs, nuis, result=result, recursion=recursion+1)
-            else:
-                result.add(pdf)
-        return result
-    
-    def unfold_constraints(self, threshold:int=50):
-        source = self.get_all_constraints()
-        return self._unfold_constraints(source, self.observables, self.nuisance_parameters, threshold=threshold)
-    
-    @staticmethod
-    def unfold_components(target, source:Set[ROOT.RooAbsArg]):
-        result = ROOT.RooArgSet()
-        if target in source:
-            result.add(target)
-        components = None
-        if isinstance(target, ROOT.RooAbsPdf):
-            components = target.getComponents()
-            components.remove(target)
-        elif isinstance(target, ROOT.RooProduct):
-            components = target.components()
-        if (components is not None) and components.getSize():
-            for component in components:
-                unfolded_components = ExtendedModel.unfold_components(component, source)
-                result.add(unfolded_components)
-        return result
-    
-    @semistaticmethod
-    def _pair_constraints(self, constraint_set:ROOT.RooArgSet, globs:ROOT.RooArgSet, nuis:ROOT.RooArgSet,
-                          base_component:bool=False):
-        nuis_list = []
-        glob_list = []
-        pdf_list = []
 
-        nuis_set = set(nuis)
-        for pdf in constraint_set:
-            target_np   = None
-            target_glob = None
-            
-            if base_component:
-                # getting base pdf component
-                components = pdf.getComponents()
-                components.remove(pdf)
-                if components.getSize():
-                    for c1 in components:
-                        for c2 in components:
-                            if c1 == c2:
-                                continue
-                            if c2.dependsOn(c1):
-                                components.remove(c1)
-                    if (components.getSize() > 1):
-                        raise RuntimeError("failed to isolate proper nuisance parameter")
-                    elif (components.getSize() == 1):
-                        target_np = components.first()
-                else:
-                    for np in nuis:
-                        if pdf.dependsOn(np):
-                            target_np = np
-                            break
-            else:
-                # getting the actual nuisance parameter
-                unfolded_nuis = ExtendedModel.unfold_components(pdf, set(nuis))
-                if not unfolded_nuis:
-                    for np in nuis:
-                        if pdf.dependsOn(np):
-                            target_np = np
-                            break
-                elif len(unfolded_nuis) > 1:
-                    raise RuntimeError("failed to isolate proper nuisance parameter")
-                else:
-                    target_np = unfolded_nuis.first()
-                    
-            if not target_np:
-                self.stdout.warning('WARNING: Could not find nuisance parameter for the constraint: {}'.format(pdf.GetName()))
-                continue
-            for glob in globs:
-                if pdf.dependsOn(glob):
-                    target_glob = glob
-                    break
-            if not target_glob:
-                self.stdout.warning('WARNING: Could not find global observable for the constraint: {}'.format(pdf.GetName()))
-                continue
-            nuis_list.append(target_np)
-            glob_list.append(target_glob)
-            pdf_list.append(pdf)
-        return nuis_list, glob_list, pdf_list
+    def unfold_constraints(self, constr_cls:Optional[List[str]]=None,
+                           recursion_limit:int=50, strip_disconnected:bool=False):
+        constraints = self.get_all_constraints()
+        unfolded_constraints = rf_ext.unfold_constraints(constraints, self.observables, self.nuisance_parameters,
+                                                         constraint_cls=constr_cls,
+                                                         recursion_limit=recursion_limit,
+                                                         strip_disconnected=strip_disconnected)
+        return unfolded_constraints
     
-    def pair_nuis_and_glob_obs(self):
-        constraint_set = self.unfold_constraints()
-        nuis_list, glob_list, _ = self._pair_constraints(constraint_set, self.global_observables, self.nuisance_parameters)
-        return nuis_list, glob_list
-
-    def pair_constraints(self, to_str=False, base_component:bool=False, sort:bool=True):
-        constraint_set = self.unfold_constraints()
-        nuisance_parameters = self.nuisance_parameters
-        if sort:
-            nuisance_parameters = RooArgSet.sort(nuisance_parameters)
-        nuis_list, glob_list, pdf_list = self._pair_constraints(constraint_set, self.global_observables, 
-                                                                nuisance_parameters,
-                                                                base_component=base_component)
-        if to_str:
-            nuis_names = [i.GetName() for i in nuis_list]
-            glob_names = [i.GetName() for i in glob_list]
-            pdf_names  = [i.GetName() for i in pdf_list]
-            size = len(nuis_names)
-            return [(pdf_names[i], nuis_names[i], glob_names[i]) for i in range(size)]
-        return nuis_list, glob_list, pdf_list
+    def pair_constraints(self, fmt:str="list", to_str=False, sort:bool=True):
+        constraint_pdfs = self.unfold_constraints()
+        return rf_ext.pair_constraints(constraint_pdfs, self.nuisance_parameters,
+                                       self.global_observables, fmt=fmt,
+                                       sort=sort, to_str=to_str)
     
     def get_constrained_nuisance_parameters(self):
         """Get the list of constrained nuisance parameters instances
         """
-        constrained_nuis, _, _ = self.pair_constraints()
+        _, constrained_nuis, _ = self.pair_constraints()
         return constrained_nuis
     
     def get_unconstrained_nuisance_parameters(self, constrained_nuis=None):
@@ -623,8 +508,8 @@ class ExtendedModel(AbstractObject):
         return unconstrained_nuis
     
     def set_constrained_nuisance_parameters_to_nominal(self, set_constant:bool=False):
-        constrained_nuis, _, constrain_pdf = self.pair_constraints()
-        for nuis, pdf in zip(constrained_nuis, constrain_pdf):
+        constrain_pdf, constrained_nuis, _ = self.pair_constraints(fmt="arglist", sort=False)
+        for pdf, nuis in zip(constrain_pdf, constrained_nuis):
             pdf_class = pdf.ClassName()
             if pdf_class in ["RooGaussian", "RooBifurGauss"]:
                 nuis.setVal(0)
@@ -636,32 +521,37 @@ class ExtendedModel(AbstractObject):
             for nuis in constrained_nuis:
                 nuis.setConstant(1)
     
-    def get_all_constraints(self, pdf:Optional[ROOT.RooAbsPdf]=None):
+    def get_all_constraints(self, pdf:Optional[ROOT.RooAbsPdf]=None, cache:bool=True):
         if pdf is None:
             pdf = self.pdf
         all_constraints = ROOT.RooArgSet()
         cache_name = "CACHE_CONSTR_OF_PDF_{}_FOR_OBS_{}".format(self.pdf.GetName(), 
                      ROOT.RooNameSet(self.data.get()).content())                 
-        constr = self.workspace.set(cache_name)
-        if constr:
+        constr_cache = self.workspace.set(cache_name)
+        if constr_cache and cache:
             # retrieve constrains from cache     
-            all_constraints.add(constr)
+            all_constraints.add(constr_cache)
         else:
             # load information needed to determine attributes from ModelConfig 
             obs = self.observables.Clone()
             nuis = self.nuisance_parameters.Clone()
             all_constraints = pdf.getAllConstraints(obs, nuis, ROOT.kFALSE)
-            
+            if (not constr_cache) and cache:
+                self.workspace.defineSet(cache_name, all_constraints)
+        """
+        # update: there is no need to
         # take care of the case where we have a product of constraint terms
         temp_all_constraints = ROOT.RooArgSet(all_constraints.GetName())
         for constraint in all_constraints:
             if constraint.IsA() == ROOT.RooProdPdf.Class():
                 buffer = ROOT.RooArgSet()
-                ExtendedModel.find_unique_prod_components(constraint, buffer)
+                ROOT.RooFitExt.unfoldProdPdfComponents(constraint, buffer)
                 temp_all_constraints.add(buffer)
             else:
                 temp_all_constraints.add(constraint)
         return temp_all_constraints
+        """
+        return all_constraints
     
     def inspect_constrained_nuisance_parameter(self, nuis, constraints):
         nuis_name = nuis.GetName()
@@ -954,7 +844,7 @@ class ExtendedModel(AbstractObject):
         return asimov_data
     
     def match_globs(self):
-        nuis_list, globs_list, pdf_list = self.pair_constraints(base_component=True)
+        pdf_list, nuis_list, globs_list = self.pair_constraints(base_component=True)
         for np, glob in zip(nuis_list, globs_list):
             glob.setVal(np.getVal())
             self.stdout.debug("set glob {} to val {}".format(glob.GetName(), glob.getVal()))
@@ -1562,25 +1452,44 @@ class ExtendedModel(AbstractObject):
             return RooArgSet.sort(variables)
         return variables
     
-    def as_dataframe(self, variable_type:Union[str, WSVariables]):
-        import pandas as pd
-        _variable_type = WSVariables.parse(variable_type)
-        # for the special case of displaying constrain terms
-        # will print the tuple of names of NP, Global Observables and Constrain Pdf
-        if _variable_type == WSVariables.CONSTRAINTS:
-            data = self.pair_constraints(to_str=True)
-            df = pd.DataFrame(data, columns =['constraint pdf', 'nuisance parameters', 'global observables'])
-            return df
+    def as_dataframe(self, variable_type:Union[str, WSVariables, ROOT.RooArgSet],
+                     asym_error:bool=False):
+        if isinstance(variable_type, ROOT.RooArgSet):
+            vcollection = variable_type
         else:
-            vcollection = self.get_variables(_variable_type)
-        df = variable_collection_to_dataframe(vcollection)
+            _variable_type = WSVariables.parse(variable_type)
+            # for the special case of displaying constrain terms
+            # will print the tuple of names of NP, Global Observables and Constrain Pdf
+            if _variable_type == WSVariables.CONSTRAINTS:
+                data = self.pair_constraints(to_str=True, fmt="dict")
+                import pandas as pd
+                df = pd.DataFrame(data).rename(columns={'pdf': 'constraint pdf',
+                                                        'nuis': 'nuisance parameters',
+                                                        'globs': 'global observables'})
+                return df
+            else:
+                vcollection = self.get_variables(_variable_type)
+        df = variable_collection_to_dataframe(vcollection, asym_error=asym_error)
         return df
     
+    def rename_dataset(self, rename_map:Dict):
+        for old_name, new_name in rename_map.items():
+            if old_name == new_name:
+                continue
+            dataset = self.workspace.data(old_name)
+            if not dataset:
+                raise RuntimeError(f"dataset \"{old_name}\" not found in the workspace")
+            check_dataset = self.workspace.data(new_name)
+            if check_dataset:
+                raise RuntimeError(f"cannot rename dataset from \"{old_name}\" to \"{new_name}\": "
+                                   f"the dataset \"{new_name}\" already exists in the workspace")
+            dataset.SetName(new_name)
+        
     def print_summary(self, items:Optional[List]=None, suppress_print:bool=False, detailed:bool=True,
                       include_patterns:Optional[List]=None, exclude_patterns:Optional[List]=None,
                       save_as:Optional[str]=None):
         if items is None:
-            items = ['workspace', 'dataset', 'snapshot', 'category', 'observable', 
+            items = ['workspace', 'dataset', 'snapshot', 'category',
                      'poi', 'detailed_nuisance_parameter']
         summary_str = ""
         # workspace
