@@ -1,4 +1,5 @@
 import os
+import copy
 import math
 import time
 import json
@@ -56,7 +57,7 @@ class Likelihood(AnalysisObject):
         return poi_val
         
     def nll_fit(self, poi_val:Optional[Union[Dict[str, float], float]]=None,
-                mode:int=2, snapshot_name:Optional[str]="currentSnapshot",
+                mode:int=2, snapshot_name:Optional[str]=AnalysisObject.kCurrentSnapshotName,
                 do_minos:bool=False):
         if not isinstance(self.poi, ROOT.RooArgSet):
             pois = ROOT.RooArgSet(self.poi)
@@ -66,8 +67,8 @@ class Likelihood(AnalysisObject):
         poi_val = self._parse_poi_val(pois, poi_val)
 
         if snapshot_name is None:
-            self.save_snapshot("tmpSnapshot")
-            snapshot_name = "tmpSnapshot"
+            self.save_snapshot(self.kTempSnapshotName)
+            snapshot_name = self.kTempSnapshotName
         
         nll_uncond = None
         nll_cond   = None
@@ -98,7 +99,7 @@ class Likelihood(AnalysisObject):
             fit_time = time.time() - start_time
             
         # unconditional nll
-        if mode in [0, 1]:
+        if mode in [0, 1, 4]:
             # restore initial parameter values after each fit
             self.load_snapshot(snapshot_name)
             # fix other pois if they are not studied
@@ -107,9 +108,15 @@ class Likelihood(AnalysisObject):
             for poi in pois:
                 poi.setConstant(0)
                 val = poi_val[poi.GetName()]
-                poi.setVal(val)
-            fit_status_uncond = self.minimizer.minimize()
-            roofit_result['uncond'] = self.minimizer.fit_result
+                if val is not None:
+                    poi.setVal(val)
+            if mode == 4:
+                self.minimizer.create_nll()
+                fit_status_uncond = None
+                roofit_result['uncond'] = None
+            else:
+                fit_status_uncond = self.minimizer.minimize()
+                roofit_result['uncond'] = self.minimizer.fit_result
             self.save_snapshot("uncondFit")
             nll_uncond = self.minimizer.nll.getVal()
             for poi in pois:
@@ -126,9 +133,13 @@ class Likelihood(AnalysisObject):
             for poi in self.model.pois:
                 poi.setConstant(1)
             for poi in pois:
-                poi.setConstant(1)
                 val = poi_val.get(poi.GetName(), poi.getVal())
-                poi.setVal(val)
+                # profiled poi
+                if val is None:
+                    poi.setConstant(0)
+                else:
+                    poi.setConstant(1)
+                    poi.setVal(val)
             fit_status_cond = self.minimizer.minimize()
             roofit_result['cond'] = self.minimizer.fit_result
             self.save_snapshot("condFit")
@@ -137,16 +148,9 @@ class Likelihood(AnalysisObject):
         
         fit_result = {}
         self.stdout.info("INFO: NLL evaluation completed with")
-        if mode in [0, 1]:
-            if len(pois) == 1:
-                poi_name = pois.first().GetName()
-                muhat = muhat[poi_name]
-                muhat_errlo = muhat_errlo[poi_name]
-                muhat_errhi = muhat_errhi[poi_name]
-                self.stdout.info("mu hat = ".rjust(15)+f"{muhat:.5f}")
-            else:
-                best_fit_str = ", ".join([f"{name}={value:.5f}" for name, value in muhat.items()])
-                self.stdout.info("best fit : ".rjust(15) + f"{best_fit_str}")
+        if mode in [0, 1, 4]:
+            best_fit_str = ", ".join([f"{name}={value:.5f}" for name, value in muhat.items()])
+            self.stdout.info("best fit : ".rjust(15) + f"{best_fit_str}")
             self.stdout.info("uncond NLL = ".rjust(15) + f"{nll_uncond:.5f}")
             fit_result['uncond_fit'] = {
                 'muhat': muhat,
@@ -157,17 +161,17 @@ class Likelihood(AnalysisObject):
                 'time': fit_time_uncond
             }
         if mode in [0, 2]:
-            if len(pois) == 1:
-                poi_name = pois.first().GetName()
-                mu = pretty_value(poi_val[poi_name])
-                self.stdout.info("mu = ".rjust(15)+f"{mu:.5f}")
-            else:
-                mu = poi_val
-                mu_str = ", ".join([f"{name}={pretty_value(value)}" for name, value in poi_val.items()])
-                self.stdout.info("mu : ".rjust(15) + f"{mu_str}")
+            mu = {name:pretty_value(value) for name, value in poi_val.items()}
+            mu_str = ", ".join([f"{name}={pretty_value(value)}" for name, value in poi_val.items()])
+            self.stdout.info("mu : ".rjust(15) + f"{mu_str}")
+            # also add the best-fit value for profiled POIs
+            for name, value in poi_val.items():
+                if value == None:
+                    muhat[name] = self.model.workspace.var(name).getVal()
             self.stdout.info("cond NLL = ".rjust(15)+f"{nll_cond:.5f}")
             fit_result['cond_fit'] = {
                 'mu': mu,
+                'muhat': muhat,
                 'nll': nll_cond,
                 'status': fit_status_cond,
                 'time': fit_time_cond
